#include "ezyaml.h"
#undef DBG
#include "lstx.h"
#include "sock.h"
#include "val.h"

int
nv_clip(double *nv, double *pa, double *pb, double *ret_a, double *ret_b)
{
	struct line nv_line, lab;
	double zero[3] = {0,0,0}, pc[3];
	int in_a, in_b;

	line_new(&nv_line, zero, nv);

	in_a = is_plane_side(&nv_line, pa);
	in_b = is_plane_side(&nv_line, pb);

	if(in_a && in_b){
		v_set(ret_a, pa);
		v_set(ret_b, pb);
		return 1;
	}
	if(!in_a && !in_b){
		return 0;
	}

	line_new_p2(&lab, pa, pb);
	cross_plane_line(&nv_line, &lab, pc);

	if(in_a){
		v_set(ret_a, pa);
		v_set(ret_b, pc);
	}else{
		v_set(ret_a, pc);
		v_set(ret_b, pb);
	}
	return 1;
}

int
eye_clip(struct lstx *wh2eye, int w, int h, double d, double *pa, double *pb, double *ret_a, double *ret_b)
{
	int n = 4, i;
	double p4[4][3] = { {0,0,-d},{w,0,-d},{w,h,-d},{0,h,-d} };
	double ta[3], tb[3];

	for(i=0; i<n; i++){
		double t[3];
		lstx_tr_p( wh2eye, L2G, p4[i], t );
		v_set( p4[i], t );
	}

	v_set(ret_a, pa);
	v_set(ret_b, pb);
	for(i=0; i<n; i++){
		double nv[3];

		v_set(ta, ret_a);
		v_set(tb, ret_b);
		cross_product_unit( p4[i], p4[(i+1)%n], nv );
		if( nv_clip(nv, ta, tb, ret_a, ret_b) == 0 ){
			return 0;
		}
	}
	return 1;
}

int
cvt(double *ps, int pn, int *idxs, int idxn, struct ax *eye2g, 
	struct lstx *wh2eye, int w, int h, double d, int *ret_xyxy)
{
	/* ret_xyxy[ idxn * 4 ] */

	double ps_[ pn * 3 ];
	double t[ idxn * 6 ];
	int tn = 0, i;

	for(i=0; i<pn; i++){
		ax_tr_p(eye2g, G2L, &ps[i*3], &ps_[i*3]);
	}

	for(i=0; i<idxn; i++){
		int ia = idxs[i*2];
		int ib = idxs[i*2+1];
		double pa[3], pb[3];		
		if( eye_clip( wh2eye, w, h, d, &ps_[ia*3], &ps_[ib*3], pa, pb ) ){
			lstx_tr_p( wh2eye, G2L, pa, &t[tn*6] );
			lstx_tr_p( wh2eye, G2L, pb, &t[tn*6+3] );
			tn++;
		}
	}

	for(i=0; i<tn; i++){
		int j = i * 6;
		*ret_xyxy++ = (int)( t[j++] );
		*ret_xyxy++ = (int)( t[j++] );
		j++;
		*ret_xyxy++ = (int)( t[j++] );
		*ret_xyxy++ = (int)( t[j++] );
	}
	return tn;
}

struct obj *
load(char *fn_yaml)
{
	FILE *fp = ( strcmp(fn_yaml, "-") == 0 ) ? stdin : fopen(fn_yaml, "r");
	struct obj *str = str_read(fp);

	if( strcmp(fn_yaml, "-") != 0 ){
		fclose(fp);
	}
	return ezyaml_load(str);
}

struct obj *
dict_get(struct obj *dic, struct obj *k)
{
	struct obj *lst_kv = dict_top_lst_kv(dic);
	int n = dict_len(dic), i;

	for(i=0; i<n; i++){
		struct obj *k_ = list_top(lst_kv);
		if( obj_eq(k_, k) ){
			return k_->next;
		}
		lst_kv = lst_kv->next;
	}
	return NULL;
}

struct obj *
dict_get_cp(struct obj *dic, char *cp)
{
	struct obj *k = str_new(cp);
	struct obj *v = dict_get(dic, k);
	obj_del(k);
	return v;
}

double
obj_get_dbl(struct obj *e)
{
	if(e->typ == OBJ_TYP_DOUBLE){
		return e->any.d;
	}
	if(e->typ == OBJ_TYP_INT){
		return e->any.i;
	}
	return 0;	
}

int
list_get_dbls(struct obj *lst, double *r, int rn)
{
	struct obj *e = list_top(lst);
	int n = list_len(lst), i;
	n = rn < n ? rn : n;
	for(i=0; i<n; i++){
		*r++ = obj_get_dbl(e);
		e = e->next;
	}
	return n;
}

double *
get_ps(struct obj *ps, int *ret_n)
{
	int n = list_len(ps), i;
	struct obj *e = list_top(ps);
	double *ret = malloc( sizeof(*ret) * n * 3 ), *p = ret;

	for(i=0; i<n; i++){
		double r[3] = {0,0,0};
		list_get_dbls(e, r, 3);
		memcpy( p, r, sizeof(r) );
		p += 3;
		e = e->next;
	}
	*ret_n = n;
	return ret;
}

int *
get_idxs(struct obj *idxs, int *ret_n)
{
	int n = list_len(idxs), i;
	struct obj *e = list_top(idxs);
	int sum = 0;
	int *ret, *rp;

	for(i=0; i<n; i++){
		sum += list_len(e) - 1;
		e = e->next;
	}

	rp = ret = malloc( sizeof(*ret) * sum * 2 );

	e = list_top(idxs);
	for(i=0; i<n; i++){
		int m = list_len(e) - 1, j;
		struct obj *f = list_top(e);
		for(j=0; j<m; j++){
			*rp++ = f->any.i;
			*rp++ = f->next->any.i;
			f = f->next;
		}
		e = e->next;
	}

	*ret_n = sum;
	return ret;
}

#define VIEW_XYOrbit		0
#define VIEW_TopDownOrtho	1
#define VIEW_N			2
int wf_view = VIEW_XYOrbit;

char wf_mode[8] = { '\0' };

double *wf_ps;
int wf_psn;
int *wf_idxs;
int wf_idxn;
double eye_p[3], eye_t[3];
double eye_p_org[3], eye_t_org[3];
double wf_bp[3], wf_bt[3];
int wf_bx, wf_by;
double wf_dz;

static void
init(char *fn_yaml, double w_deg)
{
	struct obj *dic = load(fn_yaml);

	wf_ps = get_ps( dict_get_cp(dic, "ps"), &wf_psn );
	wf_idxs = get_idxs( dict_get_cp(dic, "idxs"), &wf_idxn );
	list_get_dbls( dict_get_cp(dic, "eye_p"), eye_p, 3 );
	list_get_dbls( dict_get_cp(dic, "eye_t"), eye_t, 3 );

	v_set(eye_p_org, eye_p);
	v_set(eye_t_org, eye_t);

	wf_dz = 1.0 / tan( 0.5 * deg_to_rad(w_deg) );
}

void
wf_init(int fd)
{
	int sz;
	char *buf, *bp;
	int n;
	char *fn_yaml;
	double w_deg;

	read_n(fd, &sz, sizeof(sz));
	buf = malloc(sz);
	read_n(fd, buf, sz);

	bp = buf;
	bp = COPY_FROM_BP(bp, n);
	fn_yaml = malloc(n + 1);
	bp = COPY_SZ_FROM_BP(bp, fn_yaml, n);
	fn_yaml[n] = '\0';

	bp = COPY_FROM_BP(bp, w_deg);

	init(fn_yaml, w_deg);

	free(fn_yaml);
	free(buf);

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}

static int
get_lst(int w, int h, int *ret_xyxy)
{
	double d = wf_dz * w;

	struct x_x xlst[4] = { { TYP_AX }, { TYP_FX }, { TYP_AX }, { TYP_AX } };
	struct lstx wh2eye = { 4, xlst };
	struct line l;
	struct ax eye2g;

	ax_wh_map( &xlst[0].ax, w, h );
	fcx_new( &xlst[1].fx, wf_dz, 0 );
	ax_zoom_z( &xlst[2].ax, -1);
	ax_rot_x( &xlst[3].ax, 90);

	line_new_p2(&l, eye_p, eye_t);
	ax_new_line(&eye2g, &l);

	if( wf_view == VIEW_TopDownOrtho ){
		double a[3], deg, t[3];
		struct line l;
		struct ax ax;

		v_sub( eye_p, eye_t, a );
		unit(a, a);
		deg = rad_to_deg( acos( a[2] ) );
		cross_product_unit( a, v_z1, t );
		line_new( &l, eye_t, t );
		ax_rot( &ax, &l, deg );
		ax_compo( &eye2g, &ax, &eye2g );
	}

	return cvt(wf_ps, wf_psn, wf_idxs, wf_idxn, &eye2g, &wh2eye, w, h, d, ret_xyxy);
}

static void
get_lst_out(int fd, int w, int h, int wf_idxn)
{
	int ret_xyxy[ wf_idxn * 4 ]; /* local wf_idxn */
	int n = get_lst(w, h, ret_xyxy);
	char buf[ sizeof(n) + sizeof(int) * n * 4 ], *bp = buf;
	int sz;

	bp = COPY_TO_BP(bp, n);
	bp = COPY_SZ_TO_BP( bp, ret_xyxy, sizeof(int)*n*4 );

	sz = sizeof(buf);
	write_n(fd, &sz, sizeof(sz));
	write_n(fd, buf, sz);
}

void
wf_get_lst(int fd)
{
	int sz;
	int w, h;
	char buf[ sizeof(w) + sizeof(h) ], *bp = buf;

	read_n(fd, &sz, sizeof(sz));
	read_n(fd, buf, sz);
	bp = COPY_FROM_BP(bp, w);
	bp = COPY_FROM_BP(bp, h);

	get_lst_out(fd, w, h, wf_idxn);
}

void
wf_down(int fd)
{
	int sz, x, y, n4;;
	char mode[4];
	char buf[ sizeof(x) + sizeof(y) + sizeof(n4) + sizeof(mode) ], *bp = buf;

	read_n(fd, &sz, sizeof(sz));
	read_n(fd, buf, sizeof(buf));
	bp = COPY_FROM_BP(bp, x);
	bp = COPY_FROM_BP(bp, y);
	bp = COPY_FROM_BP(bp, n4);
	bp = COPY_SZ_FROM_BP(bp, mode, sizeof(mode));

	wf_bx = x;
	wf_by = y;
	memcpy( wf_mode, mode, sizeof(mode) );

	v_set(wf_bp, eye_p);
	v_set(wf_bt, eye_t);

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}

void
wf_up(int fd)
{
	int sz;
	read_n(fd, &sz, sizeof(sz));	

	memset(wf_mode, 0, 4);

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}

static void
drag(int x, int y, int w, int h)
{
	double w2 = w * 0.5;
	double h2 = h * 0.5;
	double rx = (x - wf_bx) / w2;
	double ry = (y - wf_by) / h2;

	if( memcmp(wf_mode, "zoom", 4) == 0 ){
		double t[3];
		v_sub( eye_t, eye_p, t );
		v_mul( t, -ry, t );
		v_add( wf_bp, t, eye_p );
	}else if( memcmp(wf_mode, "move", 4) == 0 ){
		if(wf_view == VIEW_XYOrbit){
			double vy[3], vx[3], tx[3], ty[3], mv[3];
			struct ax ax;

			v_sub( eye_t, eye_p, vy );
			vy[2] = 0;
			ax_rot_z( &ax, -90 );
			ax_tr_p( &ax, L2G, vy, vx );
			v_mul( vx, -rx, tx );
			v_mul( vy, ry, ty );
			v_add( tx, ty, mv );
			v_mul( mv, 0.1, mv );
			v_add( wf_bt, mv, eye_t );
			v_add( wf_bp, mv, eye_p );
		}else{
			double dzw = wf_dz * w;
			double rx = atan2(x - w2, dzw) - atan2(wf_bx - w2, dzw);
			double ry = atan2(y - h2, dzw) - atan2(wf_by - h2, dzw);
			struct x_x xlst[4] = { { TYP_AX }, { TYP_AX }, { TYP_AX }, { TYP_AX } };
			struct lstx lx = { 4, xlst };
			struct line l;
			struct ax *ax_p = &xlst[0].ax;
			struct ax *ax_p_rev = &xlst[3].ax;

			line_new_p2( &l, wf_bp, wf_bt );
			ax_new_line( ax_p, &l );
			ax_rev( ax_p, ax_p_rev );
			ax_rot_z( &xlst[1].ax, rx*90 );
			ax_rot_x( &xlst[2].ax, ry*90 );

			lstx_tr_p( &lx, G2L, wf_bt, eye_t );
		}
	}else if( memcmp(wf_mode, "rota", 4) == 0 ){
		struct x_x xlst[4] = { { TYP_AX }, { TYP_AX }, { TYP_AX }, { TYP_AX } };
		struct lstx lx = { 4, xlst };
		struct line l;
		struct ax *ax_t = &xlst[0].ax;
		struct ax *ax_t_rev = &xlst[3].ax;

		line_new_p2( &l, wf_bt, wf_bp );
		ax_new_line( ax_t, &l );
		ax_rev(ax_t, ax_t_rev);
		ax_rot_z( &xlst[1].ax, rx*90 );
		ax_rot_x( &xlst[2].ax, -ry*90 );

		lstx_tr_p( &lx, G2L, wf_bp, eye_p );
	}
}

void
wf_drag(int fd)
{
	int sz, x, y, w, h;
	char buf[ sizeof(x) + sizeof(y) + sizeof(w) + sizeof(h) ], *bp = buf;

	read_n(fd, &sz, sizeof(sz));	
	read_n(fd, buf, sizeof(buf));
	bp = COPY_FROM_BP(bp, x);
	bp = COPY_FROM_BP(bp, y);
	bp = COPY_FROM_BP(bp, w);
	bp = COPY_FROM_BP(bp, h);

	drag(x, y, w, h);

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}

void
wf_reset(int fd)
{
	int sz;
	read_n(fd, &sz, sizeof(sz));	

	v_set(eye_p, eye_p_org);
	v_set(eye_t, eye_t_org);

	wf_view = VIEW_XYOrbit;

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}
void
wf_turn_view(int fd)
{
	int sz;
	read_n(fd, &sz, sizeof(sz));	

	wf_view = (wf_view + 1) % VIEW_N;

	sz = 0;
	write_n(fd, &sz, sizeof(sz));
}
void
wf_in_drag(int fd)
{
	int sz, in_drag;
	read_n(fd, &sz, sizeof(sz));	

	in_drag = (wf_mode[0] != '\0');

	sz = sizeof(in_drag);
	write_n(fd, &sz, sizeof(sz));
	write_n(fd, &in_drag, sz);
}

int
main(int ac, char **av)
{
	void *funcs[] = {
		wf_init,
		wf_get_lst,
		wf_down,
		wf_up,
		wf_drag,
		wf_reset,
		wf_turn_view,
		wf_in_drag,
	};

	if(arg_is(ac, av, 1, "boot")){
		char *name = arg_s(ac, av, "name", "srv.wfrm");
		int port = arg_i(ac, av, "port", -1);
		int do_lock_func = 0;
		if(port < 0){
			char *buf = get_prop(name, NULL), *bp = buf;
			bp = COPY_FROM_BP(bp, port);
			free(buf);
		}
		sock_srv_main_loop(port, sizeof(funcs)/sizeof(*funcs), funcs, do_lock_func);
	}
	return 0;
}

/* EOF */
